/*
 * This file is part of the Code::Blocks IDE and licensed under the GNU General Public License, version 3
 * http://www.gnu.org/licenses/gpl-3.0.html
 *
 * $Revision$
 * $Id$
 * $HeadURL$
 */

#include <sdk.h>
#include <prep.h>
#ifndef CB_PRECOMP
    #include <wx/intl.h>
    #include <wx/regex.h>
    #include <wx/utils.h> // wxGetOSDirectory, wxGetEnv
#endif
#include <wx/filefn.h> // wxFileExists
#include <wx/fileconf.h> // wxFileConfig
#include <wx/tokenzr.h>
#include "compilerGNUARM.h"
#include "debuggermanager.h"

#ifdef __WXMSW__
    #include <wx/msw/registry.h>
#endif

CompilerGNUARM::CompilerGNUARM()
    : Compiler(_("GNU GCC Compiler for ARM"), _T("arm-elf-gcc"))
{
    m_Weight = 56;
    Reset();
}

CompilerGNUARM::~CompilerGNUARM()
{
    //dtor
}

Compiler * CompilerGNUARM::CreateCopy()
{
    return (new CompilerGNUARM(*this));
}

static wxString FindInstallDir(const wxString& triplet = wxT("arm-none-eabi"))
{
#ifdef __WXMSW__ // for wxRegKey
    static const wxString regloc[2] = {
        wxT("HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall"),
        wxT("HKEY_LOCAL_MACHINE\\Software\\Wow6432Node\\Microsoft\\Windows\\CurrentVersion\\Uninstall")
    };
    for (unsigned int regi = 0; regi < sizeof(regloc)/sizeof(regloc[0]); ++regi)
    {
        wxRegKey key; // defaults to HKCR
        key.SetName(regloc[regi]);
        if (!key.Open(wxRegKey::Read))
            continue;
        long li;
        wxString key_name;
        if (!key.GetFirstKey(key_name, li))
            continue;
        do {
            if (!key_name.StartsWith(wxT("GNU Tools for ARM Embedded Processor")))
                continue;
            wxRegKey key2(key, key_name);
            if (!key2.Open(wxRegKey::Read))
                continue;
            wxString uninstname;
            if (!key2.QueryValue(wxT("UninstallString"), uninstname))
                continue;
            if (uninstname.StartsWith(wxT("\"")))
                uninstname.Remove(0, 1);
            if (uninstname.EndsWith(wxT("\"")))
                uninstname.RemoveLast(1);
            if (uninstname.IsEmpty())
                continue;
            wxString dirname;
            wxFileName::SplitPath(uninstname, &dirname, 0, 0, wxPATH_NATIVE);
            if (dirname.IsEmpty())
                continue;
            wxFileName fn(dirname + wxFileName::GetPathSeparator(wxPATH_NATIVE) + wxT("bin"), triplet + wxT("-gcc.exe"), wxPATH_NATIVE);
            if (fn.IsFileExecutable())
                return dirname;
        } while (key.GetNextKey(key_name, li));
    }
#endif
#ifdef __WXGTK__
    static const wxString loc[2] = {
        wxT("/usr"),
        wxT("/usr/local")
    };
    for (unsigned int i = 0; i < sizeof(loc)/sizeof(loc[0]); ++i)
    {
        wxFileName fn(loc[i] + wxFileName::GetPathSeparator(wxPATH_NATIVE) + wxT("bin"), wxT("arm-none-eabi-gcc"), wxPATH_NATIVE);
        if (fn.IsFileExecutable())
            return loc[i];
    }
#endif
    return wxString();
}

AutoDetectResult CompilerGNUARM::AutoDetectInstallationDir()
{
    wxString sep = wxFileName::GetPathSeparator();
    static const wxString alt_triplet = wxT("arm-none-eabi");
    wxString mpath = FindInstallDir(alt_triplet);
    if (!mpath.IsEmpty())
    {
        static const wxString origprefix = wxT("arm-elf");
        m_MasterPath = mpath;
#if 0
        if (m_Programs.C.StartsWith(origprefix))
            m_Programs.C.Replace(origprefix, alt_triplet, false);
        if (m_Programs.CPP.StartsWith(origprefix))
            m_Programs.CPP.Replace(origprefix, alt_triplet, false);
        if (m_Programs.LD.StartsWith(origprefix))
            m_Programs.LD.Replace(origprefix, alt_triplet, false);
        if (m_Programs.LIB.StartsWith(origprefix))
            m_Programs.LIB.Replace(origprefix, alt_triplet, false);
#else
#ifdef __WXMSW__
        m_Programs.C = alt_triplet + wxT("-gcc.exe");
        m_Programs.CPP = alt_triplet + wxT("-g++.exe");
        m_Programs.LD = alt_triplet + wxT("-g++.exe");
        m_Programs.LIB = alt_triplet + wxT("-ar.exe");
#endif
#ifdef __WXGTK__
        m_Programs.C = alt_triplet + wxT("-gcc");
        m_Programs.CPP = alt_triplet + wxT("-g++");
        m_Programs.LD = alt_triplet + wxT("-g++");
        m_Programs.LIB = alt_triplet + wxT("-ar");
#endif
#endif
        DebuggerManager *dbgManager = Manager::Get()->GetDebuggerManager();
        const DebuggerManager::RegisteredPlugins &plugins = dbgManager->GetAllDebuggers();
        DebuggerManager::RegisteredPlugins::const_iterator it = plugins.begin();
        for (; it != plugins.end(); ++it)
            if (it->first->GetSettingsName() == wxT("gdb_debugger"))
                break;
        ConfigManagerWrapper cfg;
        bool newcfg(false);
        if (it == plugins.end()) {
            ConfigManager *mainConfig = Manager::Get()->GetConfigManager(wxT("debugger_common"));
            mainConfig->Write(wxT("/sets/gdb_debugger/conf1/name"), "Default");
            cfg = ConfigManagerWrapper(wxT("debugger_common"), wxT("/sets/gdb_debugger/conf2/values/"));
            newcfg = true;
        } else {
            const DebuggerManager::PluginData &data = it->second;
            DebuggerManager::ConfigurationVector::const_iterator itConf = data.GetConfigurations().begin();
            for (; itConf != data.GetConfigurations().end(); ++itConf)
                if ((*itConf)->GetName() == wxT("ARM"))
                    break;
            if (itConf != data.GetConfigurations().end())
                    cfg = (*itConf)->GetConfig();
            if (!cfg.IsValid()) {
                cfg = dbgManager->NewConfig(it->first, wxT("ARM"));
                newcfg = true;
            }
        }
        if (cfg.IsValid()) {
            if (newcfg) {
                ConfigManager *mainConfig = Manager::Get()->GetConfigManager(wxT("debugger_common"));
                wxString namePath = cfg.GetBasepath();
                namePath.Remove(namePath.length() - 7); // trim the "values/" from the path
                mainConfig->Write(namePath + wxT("name"), "ARM");
            }
            cfg.Write(wxT("disable_init"), true);
            cfg.Write(wxT("user_arguments"), "");
            cfg.Write(wxT("type"), 0);
            cfg.Write(wxT("init_commands"), "");
            cfg.Write(wxT("watch_args"), true);
            cfg.Write(wxT("watch_locals"), true);
            cfg.Write(wxT("watch_script_printers"), true);
            cfg.Write(wxT("catch_exceptions"), true);
            cfg.Write(wxT("eval_tooltip"), false);
            cfg.Write(wxT("add_other_search_dirs"), false);
            cfg.Write(wxT("do_not_run"), false);
            cfg.Write(wxT("disassembly_flavor"), 0);
            cfg.Write(wxT("instruction_set"), "");
#ifdef __WXMSW__
            cfg.Write(wxT("executable_path"), m_MasterPath + sep + wxT("bin") + sep + alt_triplet + wxT("-gdb.exe"));
#endif
#ifdef __WXGTK__
            cfg.Write(wxT("executable_path"), m_MasterPath + sep + wxT("bin") + sep + alt_triplet + wxT("-gdb"));
#endif
            dbgManager->RebuildAllConfigs();
        }
        m_Programs.DBGconfig = wxT("gdb_debugger:ARM");
#ifdef __WXMSW__
        AddIncludeDir(m_MasterPath + wxFILE_SEP_PATH + alt_triplet + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(m_MasterPath + wxFILE_SEP_PATH + alt_triplet + wxFILE_SEP_PATH + wxT("lib") + wxFILE_SEP_PATH + wxT("thumb") + wxFILE_SEP_PATH + wxT("v6-m"));
#endif
#ifdef __WXGTK__
        AddIncludeDir(wxT("/usr/arm-none-eabi/include"));
        AddLibDir(wxT("/usr/arm-none-eabi/lib/thumb/v6-m"));
#endif
    }
    else if (platform::windows)
    {
        // Search for GNUARM installation dir
        wxString windir = wxGetOSDirectory();
        wxFileConfig ini(_T(""), _T(""), windir + _T("/GnuARM.ini"), _T(""), wxCONFIG_USE_LOCAL_FILE | wxCONFIG_USE_NO_ESCAPE_CHARACTERS);
        // need it as const , so correct overloaded method will be selected
        wxString Programs = _T("C:\\Program Files");
        // what's the "Program Files" location
        // TO DO : support 64 bit ->    32 bit apps are in "ProgramFiles(x86)"
        //                              64 bit apps are in "ProgramFiles"
        wxGetEnv(_T("ProgramFiles"), &Programs);
        // need it as const , so correct overloaded method will be selected
        const wxString ProgramsConst = Programs + _T("\\GNUARM");
        m_MasterPath = ini.Read(_T("/InstallSettings/InstallPath"), ProgramsConst);

        if (wxFileExists(m_MasterPath + sep + _T("bin") + sep + m_Programs.C))
        {
            m_Programs.MAKE = _T("make.exe"); // we distribute "make" not "mingw32-make"
        }
    }
    else
        m_MasterPath = _T("/usr");

    wxString axsdb;
    if (platform::windows)
    {
#ifdef __WXMSW__ // for wxRegKey
        bool keyfound(false);
        {
            wxRegKey keyaxsdb;
            keyaxsdb.SetName(wxT("HKEY_LOCAL_MACHINE\\Software\\ON Semiconductor\\AXSDB"));
            if (keyaxsdb.Exists() && keyaxsdb.Open(wxRegKey::Read))
                keyfound = keyaxsdb.QueryValue(wxT("InstallDir"), axsdb);
        }
        if (!keyfound) {
            wxRegKey keyaxsdb;
            keyaxsdb.SetName(wxT("HKEY_CURRENT_USER\\Software\\ON Semiconductor\\AXSDB"));
            if (keyaxsdb.Exists() && keyaxsdb.Open(wxRegKey::Read))
                keyfound = keyaxsdb.QueryValue(wxT("InstallDir"), axsdb);
        }
#endif

        if (axsdb.IsEmpty()) {
            wxString path;
            if (wxGetEnv(wxT("PATH"), &path)) {
                wxStringTokenizer pathtok(path, wxT(";"), wxTOKEN_STRTOK);
                while (pathtok.HasMoreTokens()) {
                    wxFileName fn(pathtok.GetNextToken(), wxT(""));
                    {
                        const wxArrayString& dirs(fn.GetDirs());
                        if (!dirs.Count() || dirs.Last().CmpNoCase(wxT("bin")))
                            continue;
                    }
                    fn.RemoveLastDir();
                    if (!wxFileExists(fn.GetFullPath() + wxFILE_SEP_PATH + wxT("bin") + wxFILE_SEP_PATH + wxT("axsdb.exe")))
                        continue;
                    axsdb = fn.GetFullPath();
                    break;
                }
            }
        }
    }
    else
    {
#ifdef __WXGTK__
        axsdb = _T("/usr/share/microfoot");
#endif
    }

    if (!axsdb.IsEmpty())
    {
        // libreent does not have includes (it contains only compiler helper functions)
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libmf") + wxFILE_SEP_PATH + wxT("include"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libmf") + wxFILE_SEP_PATH + wxT("include") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libmf") + wxFILE_SEP_PATH + wxT("include") + wxFILE_SEP_PATH + wxT("CMSIS_5") + wxFILE_SEP_PATH +
                      wxT("CMSIS") + wxFILE_SEP_PATH + wxT("Core") + wxFILE_SEP_PATH + wxT("Include"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libmf") + wxFILE_SEP_PATH + wxT("include") + wxFILE_SEP_PATH + wxT("CMSIS_5") + wxFILE_SEP_PATH +
                      wxT("Device") + wxFILE_SEP_PATH + wxT("ARM") + wxFILE_SEP_PATH + wxT("ARMCM0plus") + wxFILE_SEP_PATH + wxT("Include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libmf") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libmfcrypto") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libmfcrypto") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libaxdvk2") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libaxdvk2") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libax5031") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libax5031") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libax5042") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libax5042") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libax5043") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libax5043") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libax5051") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libax5051") + wxFILE_SEP_PATH + wxT("arm"));
        AddIncludeDir(axsdb + wxFILE_SEP_PATH + wxT("libaxdsp") + wxFILE_SEP_PATH + wxT("include"));
        AddLibDir(axsdb + wxFILE_SEP_PATH + wxT("libaxdsp") + wxFILE_SEP_PATH + wxT("arm"));
    }

    // turn on GDB windows path workaround
    if (platform::windows)
    {
        m_Switches.UseFullSourcePaths = true;
    }
    return wxFileExists(m_MasterPath + wxFILE_SEP_PATH + wxT("bin") + wxFILE_SEP_PATH + m_Programs.C) ? adrDetected : adrGuessed;
} // end of AutoDetectInstallationDir
